# taken from
# https://github.com/duckdb/duckdb/blob/ff7f24fd8e3128d94371827523dae85ebaf58713/third_party/libpg_query/grammar/keywords/reserved_keywords.list
keyword_list = [
    "ALL",
    "ANALYSE",
    "ANALYZE",
    "AND",
    "ANY",
    "ARRAY",
    "AS",
    "ASC_P",
    "ASYMMETRIC",
    "BOTH",
    "CASE",
    "CAST",
    "CHECK_P",
    "COLLATE",
    "COLUMN",
    "CONSTRAINT",
    "CREATE_P",
    "DEFAULT",
    "DEFERRABLE",
    "DESC_P",
    "DISTINCT",
    "DO",
    "ELSE",
    "END_P",
    "EXCEPT",
    "FALSE_P",
    "FETCH",
    "FOR",
    "FOREIGN",
    "FROM",
    "GRANT",
    "GROUP_P",
    "HAVING",
    "QUALIFY",
    "IN_P",
    "INITIALLY",
    "INTERSECT",
    "INTO",
    "LATERAL_P",
    "LEADING",
    "LIMIT",
    "NOT",
    "NULL_P",
    "OFFSET",
    "ON",
    "ONLY",
    "OR",
    "ORDER",
    "PIVOT",
    "PIVOT_WIDER",
    "PIVOT_LONGER",
    "PLACING",
    "PRIMARY",
    "REFERENCES",
    "RETURNING",
    "SELECT",
    "SOME",
    "SYMMETRIC",
    "TABLE",
    "THEN",
    "TO",
    "TRAILING",
    "TRUE_P",
    "UNION",
    "UNIQUE",
    "UNPIVOT",
    "USING",
    "VARIADIC",
    "WHEN",
    "WHERE",
    "WINDOW",
    "WITH",
]
